from ..base import BaseOntology


class Wine(BaseOntology):
    """
    A project to define an RDF style ontology for wines and the wine-industry
    """
    ontology_id = "Wine"
    ontology_full_name = "Wine Ontology (Wine)"
    domain = "Food and Beverage"
    category = "Wine"
    version = None
    last_updated = None
    creator = None
    license = None
    format = "RDF"
    download_url = "https://github.com/UCDavisLibrary/wine-ontology"



from ..base import AutoLLM
from typing import List, Optional
from transformers import AutoModelForCausalLM, AutoTokenizer
import torch


class AutoLearnerLLM(AutoLLM):
    def __init__(self, token: str = ""):
        super().__init__()
        self.token = token

    def load(self, model_id: str = "mistralai/Mistral-7B-Instruct-v0.1", token: Optional[str] = None):
        auth_token = token if token is not None else self.token

        self.tokenizer = AutoTokenizer.from_pretrained(model_id, token=auth_token)
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.model = AutoModelForCausalLM.from_pretrained(
            model_id,
            device_map="auto",
            torch_dtype=torch.bfloat16,
            token=auth_token
        )

    def generate(self, inputs: List[str], max_new_tokens: int = 50) -> List[str]:
        encoded_inputs = self.tokenizer(inputs, return_tensors="pt",
                                        padding=True).to(self.model.device)
        outputs = self.model.generate(
            **encoded_inputs,
            max_new_tokens=max_new_tokens,
            pad_token_id=self.tokenizer.eos_token_id,
            temperature=0.1
        )
        return [self.tokenizer.decode(output, skip_special_tokens=True) for output in outputs]



from evaluate import aggregate_metrics
from metrics import calculate_term_typing_metrics, calculate_taxonomy_metrics, calculate_non_taxonomy_metrics
from visualisations import plot_model_comparison, plot_type_analysis, plot_precision_recall_distribution

__all__ = [
    'aggregate_metrics',
    'calculate_term_typing_metrics',
    'calculate_taxonomy_metrics',
    'calculate_non_taxonomy_metrics',
    'plot_model_comparison',
    'plot_type_analysis',
    'plot_precision_recall_distribution'
]



import os
import re
import json
from abc import ABC
import concurrent.futures
from typing import List, Tuple, Any, Set, Optional
from rdflib import Graph, OWL, URIRef, RDFS, RDF
import networkx as nx
from huggingface_hub import hf_hub_download

from ..data_structure import (OntologyData, TermTyping, TaxonomicRelation, NonTaxonomicRelation,
                              TypeTaxonomies, NonTaxonomicRelations)
from .. import logger


class BaseOntology(ABC):
    """Base class for ontology processing"""
    ontology_id: str = None
    ontology_full_name: str = None
    domain: str = None
    category: str = None
    version: str = None
    last_updated: str = None
    creator: str = None
    license: str = None
    format: str = None
    download_url = None
    loaded_from_huggingface = False
    loaded_from_local = False

    def __init__(self, language: str = 'en', base_dir: Optional[str] = None):
        """Initialize the ontology"""
        self.rdf_graph = None
        self.nx_graph = None
        self.language = language
        self.base_dir = base_dir

    @staticmethod
    def from_huggingface(ontology_id: str, domain: str, format: str) -> str:
        """Download an ontology file from a Hugging Face repository."""
        if not ontology_id or not domain or not format:
            raise ValueError("Ontology ID, domain, and format must be defined")

        ontology_domain = domain.lower().replace(' ', '_')
        repo_id = f"SciKnowOrg/ontolearner-{ontology_domain}"
        ontology_id = ontology_id.lower()
        filename = f"{ontology_id}/{ontology_id}.{format.lower()}"

        try:
            file_path = hf_hub_download(repo_id=repo_id, filename=filename, repo_type="dataset")
            return file_path
        except Exception as e:
            logger.error(f"Error downloading ontology from Hugging Face: {str(e)}")
            raise

    @staticmethod
    def from_local(path: str) -> str:
        """Specify a local ontology file path."""
        if not os.path.exists(path):
            raise FileNotFoundError(f"Ontology file not found at {path}")
        return path

    def load(self, path: Optional[str] = None) -> None:
        """Load an ontology from a local file or Hugging Face repository."""
        self.loaded_from_huggingface = False
        self.loaded_from_local = False

        try:
            self.rdf_graph = Graph()

            if path:
                # Use provided path as local file
                file_path = self.from_local(path)
                self.loaded_from_local = True
            else:
                # Download from HuggingFace
                file_path = self.from_huggingface(
                    ontology_id=self.ontology_id,
                    domain=self.domain,
                    format=self.format
                )
                self.loaded_from_huggingface = True

            self._load(file_path)
            if len(self.rdf_graph) == 0:
                raise ValueError("Loaded ontology contains no triples")
            logger.info(f"Successfully loaded ontology with {len(self.rdf_graph)} triples")
        except Exception as e:
            logger.error(f"Error loading ontology: {str(e)}")
            raise

    def _load(self, path: str, visited: Optional[set] = None) -> None:
        if visited is None:
            visited = set()
        if path in visited:
            return
        visited.add(path)

        self.rdf_graph.parse(path)

        if not self.contains_imports():
            return

        # Process owl:imports
        for ontology in self.rdf_graph.subjects(RDF.type, OWL.Ontology):
            for import_def in self.rdf_graph.objects(ontology, OWL.imports):
                import_uri = self._resolve_import_def(import_def)

                if import_uri:
                    try:
                        self._load(import_uri, visited)
                    except Exception as e:
                        logger.error(f"Failed to load import {import_uri}: {str(e)}")
                else:
                    logger.warning(f"Could not resolve import: {import_def}")


    def contains_imports(self) -> bool:
        """Hook: Check if the ontology contains imports."""
        return False

    def _resolve_import_def(self, uri: URIRef) -> Optional[str]:
        """Resolve imported URIs to local file paths."""
        uri_str = str(uri)

        # Handle OBO PURLs
        if uri_str.startswith("http://purl.obolibrary.org/obo/"):
            path = uri_str.replace("http://purl.obolibrary.org/obo/", "")
            if self.base_dir:
                local_path = os.path.join(self.base_dir, path)
                if os.path.exists(local_path):
                    return local_path
            # Fallback to HTTP if local file not found
            logger.info(f"Fetching import from HTTP: {uri_str}")
            return uri_str
        # Handle other HTTP URIs
        elif uri_str.startswith("http://") or uri_str.startswith("https://"):
            logger.info(f"Fetching import from HTTP: {uri_str}")
            return uri_str
        # Handle file:// URIs
        elif uri_str.startswith("file:///"):
            file_path = uri_str[8:]
        elif uri_str.startswith("file://"):
            file_path = uri_str[7:]
        else:
            file_path = uri_str

        # Convert path separators to Unix style
        file_path = file_path.replace('\\', '/')

        # Handle Windows drive letter
        if ':' in file_path and os.name == 'nt':
            file_path = file_path.split(':', 1)[1]

        if self.base_dir:
            resolved_path = os.path.join(self.base_dir, file_path.lstrip('/'))
            if os.path.exists(resolved_path):
                return resolved_path

        return None

    def extract(self, reinforce_extraction: bool = False) -> OntologyData:
        """Extract all information from all the three functions below."""
        if not (self.loaded_from_local or self.loaded_from_huggingface):
            raise ValueError("Ontology must be loaded before extraction")

        if self.loaded_from_local:
            logger.info(f"Extracting from local source for {self.ontology_id}")
            return self._extract_from_local()
        else:
            return self._extract_from_huggingface(reinforce_extraction)

    def _extract_from_local(self) -> OntologyData:
        """Extract data from local ontology source."""
        with concurrent.futures.ThreadPoolExecutor() as executor:
            term_typings_future = executor.submit(self.extract_term_typings)
            taxonomies_future = executor.submit(self.extract_type_taxonomies)
            non_taxonomic_future = executor.submit(self.extract_type_non_taxonomic_relations)

            term_typings = term_typings_future.result()
            types, taxonomies = taxonomies_future.result()
            types_nt, relations, non_taxonomies = non_taxonomic_future.result()

            return OntologyData(
                term_typings=term_typings,
                type_taxonomies=TypeTaxonomies(
                    types=types,
                    taxonomies=taxonomies
                ),
                type_non_taxonomic_relations=NonTaxonomicRelations(
                    types=types_nt,
                    relations=relations,
                    non_taxonomies=non_taxonomies
                )
            )

    def _extract_from_huggingface(self, reinforce_extraction: bool) -> Optional[OntologyData]:
        """Extract data from HuggingFace, with optional reinforcement."""
        ontology_domain = self.domain.lower().replace(' ', '_')
        repo_id = f"SciKnowOrg/ontolearner-{ontology_domain}"
        ontology_id = self.ontology_id.lower()
        original_huggingface = self.loaded_from_huggingface
        original_local = self.loaded_from_local

        if reinforce_extraction:
            logger.info(f"Attempting reinforced extraction from HuggingFace for {self.ontology_id}")
            try:
                self.loaded_from_huggingface = False
                self.loaded_from_local = True
                result = self.extract(reinforce_extraction=False)
                return result
            except Exception as e:
                logger.error(f"Reinforced extraction failed for {self.ontology_id}: {str(e)}")
            finally:
                self.loaded_from_huggingface = original_huggingface
                self.loaded_from_local = original_local

        logger.info(f"Loading pre-extracted data from HuggingFace for {self.ontology_id}")
        try:
            term_typings_path = hf_hub_download(repo_id=repo_id,
                                                filename=f"{ontology_id}/term_typings.json",
                                                repo_type="dataset")
            taxonomies_path = hf_hub_download(repo_id=repo_id,
                                              filename=f"{ontology_id}/type_taxonomies.json",
                                              repo_type="dataset")
            non_taxonomic_path = hf_hub_download(repo_id=repo_id,
                                                 filename=f"{ontology_id}/type_non_taxonomic_relations.json",
                                                 repo_type="dataset")

            with open(term_typings_path, 'r') as f:
                term_typings = json.load(f)
            with open(taxonomies_path, 'r') as f:
                type_taxonomies = json.load(f)
            with open(non_taxonomic_path, 'r') as f:
                type_non_taxonomic_relations = json.load(f)

            return OntologyData(
                term_typings=term_typings,
                type_taxonomies=type_taxonomies,
                type_non_taxonomic_relations=type_non_taxonomic_relations
            )
        except Exception as e:
            raise ValueError(f"Failed to extract ontology data from HuggingFace: {str(e)}") from e

    @staticmethod
    def is_valid_label(label: str) -> Any:
        invalids = ['root', 'thing']
        if label.lower() in invalids:
            return None
        return label

    def get_label(self, uri: str) -> str:
        """
        Extracts the label for a given URI in the specified language from the RDF graph.
        If no valid label is found, returns None.
        """
        entity = URIRef(uri)
        labels = list(self.rdf_graph.objects(subject=entity, predicate=RDFS.label))
        for label in labels:
            if hasattr(label, 'language') and label.language == self.language:
                return self.is_valid_label(str(label))
        if labels:
            first_label = str(labels[0])
            if len(first_label) > 3 and not first_label.startswith("http"):
                return self.is_valid_label(first_label)
        if "#" in uri:
            local_name = uri.split("#")[-1]
        elif "/" in uri:
            local_name = uri.split("/")[-1]
        else:
            local_name = uri
        label = self.is_valid_label(local_name)
        if not label:
            logger.warning(f"No valid label for URI: {uri}")
        return label

    def build_graph(self) -> None:
        """
        Build NetworkX graph from RDF data.

        This method should be implemented by each specific ontology class
        to handle their unique graph structure.
        """
        self.nx_graph = nx.DiGraph()

        if not self.rdf_graph:
            logger.info("Loading from HuggingFace")
            ontology_domain = self.domain.lower().replace(' ', '_')
            repo_id = f"SciKnowOrg/ontolearner-{ontology_domain}"
            ontology_id = self.ontology_id.lower()
            file_path = hf_hub_download(
                repo_id=repo_id,
                filename=f"{ontology_id}/{ontology_id}.{self.format.lower()}",
                repo_type="dataset"
            )

            self.load(file_path)

        for subject, predicate, obj in self.rdf_graph:
            subject_label = self.get_label(str(subject))
            object_label = self.get_label(str(obj))
            predicate_label = self.get_label(str(predicate))
            if subject_label and object_label and predicate_label:
                self.nx_graph.add_node(subject_label)
                self.nx_graph.add_node(object_label)
                self.nx_graph.add_edge(subject_label, object_label, label=predicate_label)

    # ------------------- Term Typings -------------------
    def extract_term_typings(self) -> List[TermTyping]:
        """
        Extract term-to-type mappings (e.g., instances of classes).
        """
        term_typings = []
        for class_uri in self._get_relevant_classes():
            for instance in self._get_instances_for_class(class_uri):
                term = self.get_label(uri=str(instance))
                types = self.get_label(uri=str(class_uri))
                # Filter out anonymous class identifiers
                if (term and types and
                        not self._is_anonymous_id(term) and
                        not self._is_anonymous_id(types)):
                    term_typings.append(TermTyping(term=term, types=[types]))
        logger.debug(f"Extracted {len(term_typings)} term typings for the Ontology.")
        return term_typings

    def _get_relevant_classes(self) -> Set[URIRef]:
        """Hook: Define which classes to process (default: all classes)."""
        return (set(self.rdf_graph.subjects(RDF.type, RDFS.Class)) |
                set(self.rdf_graph.subjects(RDF.type, OWL.Class)))

    def _get_instances_for_class(self, class_uri: URIRef) -> Set[URIRef]:
        """Hook: Get instances of a class (default: direct instances)."""
        return set(self.rdf_graph.subjects(RDF.type, class_uri))

    # ------------------- Taxonomy Extraction -------------------
    def extract_type_taxonomies(self) -> Tuple[List[str], List[TaxonomicRelation]]:
        """
        Extract taxonomy from the ontology
        """
        types, taxonomies = [], []
        subclasses = self.rdf_graph.subjects(predicate=RDFS.subClassOf)
        for subclass in subclasses:
            parent_classes = self.rdf_graph.objects(subject=subclass, predicate=RDFS.subClassOf)
            subclass_label = self.get_label(str(subclass))
            for parent in parent_classes:
                parent_label = self.get_label(uri=str(parent))
                if (subclass_label and parent_label and
                        not self._is_anonymous_id(subclass_label) and
                        not self._is_anonymous_id(parent_label)):
                    types.append(subclass_label)
                    types.append(parent_label)
                    taxonomies.append(TaxonomicRelation(parent=parent_label, child=subclass_label))
        types = list(set(types))
        logger.debug(f"Extracted {len(taxonomies)} taxonomic relations for the Ontology.")
        return types, taxonomies

    # ------------------- Non-Taxonomic Relations -------------------
    def extract_type_non_taxonomic_relations(self) -> Tuple[List[str], List[str], List[NonTaxonomicRelation]]:
        """
        Extract non-taxonomic relations from the ontology.
        """
        types_set = set()
        relations_set = set()
        non_taxonomic_pairs: List[NonTaxonomicRelation] = []

        for s, p, o in self.rdf_graph:
            if self._is_valid_non_taxonomic_triple(s, p, o):
                head = self.get_label(str(s))
                tail = self.get_label(str(o))
                relation = self.get_label(str(p))

                # Filter out anonymous class identifiers
                if (head and tail and relation and
                        not self._is_anonymous_id(head) and
                        not self._is_anonymous_id(tail)):
                    non_taxonomic_pairs.append(
                        NonTaxonomicRelation(head=head, tail=tail, relation=relation)
                    )
                    types_set.update([head, tail])
                    relations_set.add(relation)

        types = sorted(types_set)
        relations = sorted(relations_set)
        logger.debug(f"Extracted {len(non_taxonomic_pairs)} non-taxonomic relations for the Ontology.")
        return types, relations, non_taxonomic_pairs

    def _is_valid_non_taxonomic_triple(self, s: URIRef, p: URIRef, o: URIRef) -> bool:
        """Validate non-taxonomic relations between named classes (URIRefs only)."""
        return (
            self.check_if_class(s) and
            self.check_if_class(o) and
            p != RDFS.subClassOf
        )

    def check_if_class(self, entity):
        """Check if an entity is a class (i.e., rdf:type rdfs:Class or owl:Class)."""
        for _, _, obj in self.rdf_graph.triples((entity, RDF.type, None)):
            if obj in (RDFS.Class, OWL.Class):
                return True
        return False

    @staticmethod
    def _is_anonymous_id(label: str) -> bool:
        """Check if a label represents an anonymous class identifier."""
        if not label:
            return True

        # Common RDF/OWL blank node patterns
        if label.startswith('_:'):  # Standard RDF blank node notation
            return True
        if label.startswith('genid-'):  # Common OWL tools like Protégé
            return True
        if label.startswith('nodeID://'):  # Some RDF serializers
            return True

        # Numeric patterns
        if re.match(r'^N[0-9]+$', label):
            return True
        if re.match(r'^_[0-9]+$', label):
            return True
        if re.match(r'^c_[0-9]+$', label):
            return True
        if re.match(r'^BFO_[0-9]+$', label):
            return True
        if re.match(r'^IAO_[0-9]+$', label):
            return True
        if re.match(r'^OBI_[0-9]+$', label):
            return True
        if re.match(r'^FIX_[0-9]+$', label):
            return True
        if re.match(r'^REX_[0-9]+$', label):
            return True
        if re.match(r'^UO_[0-9]+$', label):
            return True
        if re.match(r'^MS_[0-9]+$', label):
            return True
        if re.match(r'^AFRL_[0-9]+$', label):
            return True
        if re.match(r'^AFFN_[0-9]+$', label):
            return True
        if re.match(r'^AFE_[0-9]+$', label):
            return True
        if re.match(r'^AFQ_[0-9]+$', label):
            return True
        if re.match(r'^AFP_[0-9]+$', label):
            return True
        if re.match(r'^AFM_[0-9]+$', label):
            return True
        if re.match(r'^AFC_[0-9]+$', label):
            return True
        if re.match(r'^ENVO_[0-9]+$', label):
            return True
        if re.match(r'^AFR_[0-9]+$', label):
            return True

        # Hexadecimal patterns
        if re.match(r'^N[0-9a-f]{32}$', label, re.IGNORECASE):
            return True
        if re.match(r'^n[0-9a-f]+$', label, re.IGNORECASE):
            return True
        if re.match(r'^b[0-9a-f]+$', label, re.IGNORECASE):
            return True
        if re.match(r'^c_[0-9a-f]+$', label, re.IGNORECASE):
            return True

        # UUID patterns
        if re.match(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', label, re.IGNORECASE):
            return True

        # Auto-generated node patterns
        if re.match(r'^node_[0-9a-f_]+$', label, re.IGNORECASE):
            return True
        if re.match(r'^auto_gen_[0-9a-zA-Z]+$', label):
            return True
        if re.match(r'^blank[0-9]+$', label):
            return True

        # Tool-specific patterns
        if label.startswith('ARQ'):  # Apache Jena ARQ
            return True
        if label.startswith('jena-'):  # Apache Jena
            return True
        if label.startswith('bnode'):  # Some RDF tools
            return True

        # Image and label patterns
        if re.match(r'^img_', label):  # Any string starting with img_
            return True
        if re.match(r'^xl_', label):   # Any string starting with xl_
            return True
        if re.match(r'^xl-', label):   # Any string starting with xl-
            return True

        # SKOS Collection patterns
        if re.match(r'^skosCollection_[0-9a-f]+$', label):
            return True

        if re.match(r'^PMD_[0-9]+$', label):
            return True

        return False



from ..base.text2onto import BaseText2OntoDataset


class OLLMWikipedia(BaseText2OntoDataset):
    """
    OLLM Wikipedia is a synthetic ontology generated from Wikipedia articles.
    It is used to evaluate the performance of ontology learning algorithms.
    The ontology is generated from the Wikipedia dump and contains concepts and their relationships.

    This class processes OLLM Wikipedia using default textual processing methods.
    """
    dataset_id: str = "OLLMWikipedia"
    dataset_full_name: str = "OLLM Wikipedia Ontology"



import random
from typing import Tuple, List, Set
import logging
from sklearn.model_selection import train_test_split as sklearn_split

from ontolearner.data_structure import OntologyData, TermTyping, TaxonomicRelation, NonTaxonomicRelation

logger = logging.getLogger(__name__)


def term_typing_split(term_typings: List[TermTyping], test_size: float = 0.2, random_state: int = 42) \
        -> Tuple[List[TermTyping], List[TermTyping]]:
    """
    Split term typing data ensuring similar type distribution in train and test
    """
    # Get type frequencies
    type_counter = {}
    term_to_types = {}

    for tt in term_typings:
        term_to_types[tt.term] = tt.types
        for t in tt.types:
            type_counter[t] = type_counter.get(t, 0) + 1

    # Group terms by their most frequent type
    term_by_primary_type = {}
    for term, types in term_to_types.items():
        if not types:
            continue
        # Find the least common type for better stratification
        primary_type = min(types, key=lambda t: type_counter.get(t, 0))
        if primary_type not in term_by_primary_type:
            term_by_primary_type[primary_type] = []
        term_by_primary_type[primary_type].append(term)

    # Split terms within each type group
    train_terms = set()
    test_terms = set()

    for type_name, terms in term_by_primary_type.items():
        if len(terms) < 2:  # If only one term has this type, put in train
            train_terms.update(terms)
            continue

        n_test = max(1, int(len(terms) * test_size))
        type_test_terms = set(random.Random(random_state).sample(terms, n_test))
        type_train_terms = set(terms) - type_test_terms

        train_terms.update(type_train_terms)
        test_terms.update(type_test_terms)

    # Create the final train/test lists
    train_typings = [tt for tt in term_typings if tt.term in train_terms]
    test_typings = [tt for tt in term_typings if tt.term in test_terms]

    logger.info(f"Term typing split: {len(train_typings)} train, {len(test_typings)} test")
    return train_typings, test_typings


def taxonomy_split(taxonomies: List[TaxonomicRelation], train_terms: Set[str] = None,
                   test_size: float = 0.2, random_state: int = 42) \
        -> Tuple[List[TaxonomicRelation], List[TaxonomicRelation]]:
    """
    Split taxonomy relations ensuring no term leakage from train to test
    """
    train_terms = train_terms or set()
    test_candidate_relations = []
    train_relations = []

    # First pass: separate relations involving training terms
    for rel in taxonomies:
        if rel.parent in train_terms or rel.child in train_terms:
            train_relations.append(rel)
        else:
            test_candidate_relations.append(rel)

    # Second pass: split remaining relations
    if test_candidate_relations:
        n_test = min(len(test_candidate_relations),
                     max(1, int(len(taxonomies) * test_size)))

        # Use sklearn's function for this step
        train_idx, test_idx = sklearn_split(
            range(len(test_candidate_relations)),
            test_size=n_test/len(test_candidate_relations),
            random_state=random_state
        )

        for idx in train_idx:
            train_relations.append(test_candidate_relations[idx])

        test_relations = [test_candidate_relations[idx] for idx in test_idx]
    else:
        # If all relations involve training terms, do a random split
        n_test = max(1, int(len(train_relations) * test_size))
        random.Random(random_state).shuffle(train_relations)
        test_relations = train_relations[-n_test:]
        train_relations = train_relations[:-n_test]

    logger.info(f"Taxonomy split: {len(train_relations)} train, {len(test_relations)} test")
    return train_relations, test_relations


def non_taxonomic_re_split(
        non_taxonomies: List[NonTaxonomicRelation],
        train_terms: Set[str] = None,
        test_size: float = 0.2,
        random_state: int = 42
) -> Tuple[List[NonTaxonomicRelation], List[NonTaxonomicRelation]]:
    """
    Split non-taxonomy relations ensuring no term leakage from train to test
    """
    train_terms = train_terms or set()
    test_candidate_relations = []
    train_relations = []

    # First pass: separate relations involving training terms
    for rel in non_taxonomies:
        if rel.head in train_terms or rel.tail in train_terms:
            train_relations.append(rel)
        else:
            test_candidate_relations.append(rel)

    # Second pass: ensure relation type distribution
    relation_types = {}
    for rel in test_candidate_relations:
        if rel.relation not in relation_types:
            relation_types[rel.relation] = []
        relation_types[rel.relation].append(rel)

    test_relations = []

    # For each relation type, take a proportional number for test
    for rel_type, rels in relation_types.items():
        n_test = max(1, int(len(rels) * test_size))
        sampled_test = random.Random(random_state).sample(rels, min(n_test, len(rels)))
        test_relations.extend(sampled_test)

        # Remove the test relations from candidates
        for rel in sampled_test:
            test_candidate_relations.remove(rel)

    # Add remaining candidates to train
    train_relations.extend(test_candidate_relations)

    logger.info(f"Non-taxonomy split: {len(train_relations)} train, {len(test_relations)} test")
    return train_relations, test_relations


def train_test_split(data: OntologyData, test_size: float = 0.2, random_state: int = 42) \
        -> Tuple[OntologyData, OntologyData]:
    """
    Create train/test split of ontology data, ensuring consistent term assignment
    """
    # First split term typing to establish term assignment
    train_typings, test_typings = term_typing_split(
        data.term_typings, test_size, random_state
    )

    # Get sets of training and test terms
    train_terms = {tt.term for tt in train_typings}
    test_terms = {tt.term for tt in test_typings}

    # Split taxonomic relations considering term assignment
    train_taxonomies, test_taxonomies = taxonomy_split(
        data.type_taxonomies.taxonomies,
        train_terms,
        test_size,
        random_state
    )

    # Update term sets based on taxonomy split
    train_terms.update({rel.parent for rel in train_taxonomies})
    train_terms.update({rel.child for rel in train_taxonomies})
    test_terms.update({rel.parent for rel in test_taxonomies})
    test_terms.update({rel.child for rel in test_taxonomies})

    # Split non-taxonomic relations considering term assignment
    train_non_taxonomies, test_non_taxonomies = non_taxonomic_re_split(
        data.type_non_taxonomic_relations.non_taxonomies,
        train_terms,
        test_size,
        random_state
    )

    # Create train and test ontology data objects
    train_data = OntologyData(
        term_typings=train_typings,
        type_taxonomies=data.type_taxonomies.__class__(
            types=data.type_taxonomies.types,
            taxonomies=train_taxonomies
        ),
        type_non_taxonomic_relations=data.type_non_taxonomic_relations.__class__(
            types=data.type_non_taxonomic_relations.types,
            relations=data.type_non_taxonomic_relations.relations,
            non_taxonomies=train_non_taxonomies
        )
    )

    test_data = OntologyData(
        term_typings=test_typings,
        type_taxonomies=data.type_taxonomies.__class__(
            types=data.type_taxonomies.types,
            taxonomies=test_taxonomies
        ),
        type_non_taxonomic_relations=data.type_non_taxonomic_relations.__class__(
            types=data.type_non_taxonomic_relations.types,
            relations=data.type_non_taxonomic_relations.relations,
            non_taxonomies=test_non_taxonomies
        )
    )

    # Log the split summary
    logger.info("Ontology data split complete:")
    logger.info(f"  Term typing: {len(train_typings)} train, {len(test_typings)} test")
    logger.info(f"  Taxonomy: {len(train_taxonomies)} train, {len(test_taxonomies)} test")
    logger.info(f"  Non-taxonomy: {len(train_non_taxonomies)} train, {len(test_non_taxonomies)} test")

    return train_data, test_data



import re
from typing import Any, List

from ..base import AutoLearner
from ..data_structure import OntologyData


class AutoRAGLearner(AutoLearner):
    def __init__(self, learner_retriever: Any, learner_llm: Any, prompting: Any):
        super().__init__()
        self.retriever = learner_retriever
        self.llm = learner_llm
        self.prompting = prompting

    def load(self, retriever_id: str, llm_id: str):
        self.retriever.load(retriever_id)
        self.llm.load(llm_id)

    @staticmethod
    def _prepare_documents(data: OntologyData, task: str) -> List[str]:
        documents = []
        if task == "term-typing":
            for tt in data.term_typings:
                documents.append(f"Term: {tt.term}\nTypes: {', '.join(tt.types)}")
        elif task == "taxonomy-discovery":
            for tr in data.type_taxonomies.taxonomies:
                documents.append(f"Parent: {tr.parent}\nChild: {tr.child}\nRelation: is-a")
        elif task == "task-non-taxonomic-relations":
            for nr in data.type_non_taxonomic_relations.non_taxonomies:
                documents.append(f"Head: {nr.head}\nRelation: {nr.relation}\nTail: {nr.tail}")
        return documents

    def fit(self, train_data: OntologyData, task: str):
        documents = self._prepare_documents(train_data, task)
        self.retriever.index(documents)
        return self

    def predict(self, eval_data: Any, task: str) -> List[str]:
        if task == "term-typing":
            term = eval_data
            query = term
            prompt_data = {"term": term}
        elif task == "taxonomy-discovery":
            parent, child = eval_data
            query = f"{parent} {child}"
            prompt_data = {"parent": parent, "child": child}
        elif task == "task-non-taxonomic-relations":
            head, tail = eval_data
            query = f"{head} {tail}"
            prompt_data = {"head": head, "tail": tail}
        else:
            raise ValueError(f"Task {task} not supported")

        context = "\n".join(self.retriever.retrieve(query, top_k=5))

        prompting = self.prompting.format(term=prompt_data["term"], context=context)

        raw_response = self.llm.generate([prompting])[0]

        if task == "term-typing":
            match = re.search(r'\[(.*?)\]', raw_response)
            if match:
                types_str = match.group(1)
                types = [t.strip().strip("'\"") for t in types_str.split(',')]
                return types
            return [raw_response.strip()]

        return [raw_response]



__version__ = "1.1.0"

import logging

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)

logger = logging.getLogger("ontolearner")

from ontolearner import base, data_structure, ontology, text2onto, utils, tools, learner
from .processor import Processor
from .learner_pipeline import LearnerPipeline

__all__ = [
    "base",
    "ontology",
    "text2onto",
    "learner",
    "LearnerPipeline",
    "Processor",
    "tools",
    "data_structure",
    "utils",
]



from pydantic import BaseModel, Field
from typing import Dict


class TopologyMetrics(BaseModel):
    """Metrics describing the structural properties of the ontology graph"""
    # Basic graph metrics
    total_nodes: int = Field(..., description="Total number of nodes (concepts) in the ontology", ge=0)
    total_edges: int = Field(..., description="Total number of edges (relationships) between nodes", ge=0)
    num_root_nodes: int = Field(..., description="Number of root concepts", ge=0)
    num_leaf_nodes: int = Field(..., description="Number of leaf concepts", ge=0)

    # Knowledge coverage
    num_classes: int = Field(..., description="Number of classes in the ontology", ge=0)
    num_properties: int = Field(..., description="Number of properties (object/data) in the ontology", ge=0)
    num_individuals: int = Field(..., description="Number of individuals in the ontology", ge=0)

    # Hierarchical metrics
    max_depth: int = Field(..., description="Maximum depth of the concept hierarchy", ge=0)
    min_depth: int = Field(..., description="Minimum depth of the concept hierarchy", ge=0)
    avg_depth: float = Field(..., description="Average depth of nodes", ge=0.0)
    depth_variance: float = Field(..., description="Variance of node depths", ge=0.0)

    # Breadth metrics
    max_breadth: int = Field(..., description="Largest number of nodes at a level", ge=0)
    min_breadth: int = Field(..., description="Smallest number of nodes at a level", ge=0)
    avg_breadth: float = Field(..., description="Average nodes per level", ge=0.0)
    breadth_variance: float = Field(..., description="Variance of nodes per level", ge=0.0)


class DatasetMetrics(BaseModel):
    """Metrics for evaluating the quality and characteristics of generated datasets"""
    num_term_types: int = Field(..., description="Number of term-type pairs in the dataset", ge=0)
    num_taxonomic_relations: int = Field(..., description="Number of taxonomic relations", ge=0)
    num_non_taxonomic_relations: int = Field(..., description="Number of non-taxonomic relations", ge=0)
    avg_terms: float = Field(..., description="Average number of terms per type", ge=0.0)


class OntologyMetrics(BaseModel):
    """Complete metrics collection for an ontology"""
    name: str = Field(..., description="Name of the ontology")
    topology: TopologyMetrics = Field(..., description="Structural topology metrics")
    dataset: DatasetMetrics = Field(..., description="Dataset metrics")


class BenchmarkMetrics(BaseModel):
    """Aggregate metrics across multiple ontologies"""
    # Overview
    total_ontologies: int = Field(..., description="Number of ontologies analyzed")
    ontology_metrics: Dict[str, OntologyMetrics] = Field(..., description="Individual metrics for each ontology")
    # Averages
    avg_topology: TopologyMetrics = Field(..., description="Average topology metrics across ontologies")
    avg_dataset: DatasetMetrics = Field(..., description="Average dataset metrics")
    # Variation
    topology_std: Dict[str, float] = Field(..., description="Standard deviation of topology metrics")
    relationship_std: Dict[str, float] = Field(..., description="Standard deviation of relationship metrics")
    dataset_std: Dict[str, float] = Field(..., description="Standard deviation of dataset metrics")



from .agriculture import * # NOQA
from .arts_humanities import * # NOQA
from .upper_ontologies import * # NOQA
from .biology import * # NOQA
from .chemistry import * # NOQA
from .education import * # NOQA
from .ecology_environment import * # NOQA
from .events import * # NOQA
from .finance import * # NOQA
from .food_beverage import * # NOQA
from .general import * # NOQA
from .geography import * # NOQA
from .medicine import * # NOQA
from .industry import * # NOQA
from .law import * # NOQA
from .library_cultural_heritage import * # NOQA
from .material_science_engineering import * # NOQA
from .news_media import * # NOQA
from .scholarly_knowledge import * # NOQA
from .social_sciences import * # NOQA
from .units_measurements import * # NOQA
from .web import * # NOQA



from pathlib import Path
from typing import Union, Dict
import pandas as pd
from jinja2 import Template
import time

from .base import BaseOntology
from .data_structure import OntologyMetrics, OntologyData
from .tools import Analyzer
from .utils import io
from . import logger

class Processor:
    """
    Handles the complete ontology processing pipeline including:
    loading, extraction, analysis, and saving of results.
    """
    def __init__(self, datasets_dir: Path, templates_dir: Path, benchmark_dir: Path,  metrics_dir: Path, analyzer_class: type[Analyzer] = Analyzer):
        self.datasets_dir = datasets_dir
        self.templates_dir = templates_dir
        self.benchmark_dir = benchmark_dir
        self.metrics_dir = metrics_dir
        self.analyzer_class = analyzer_class
        self.all_metrics: Dict[str, dict] = {}
        self.doc_template_path = templates_dir / "ontology.rst"
        self.doc_template = Template(self.doc_template_path.read_text())

    def process_ontology(self, ontology: BaseOntology, ontology_path: Union[str, Path]) \
            -> OntologyMetrics:
        """Process a single ontology through the complete pipeline."""
        start_time = time.time()

        try:
            logger.info(f"Processing {ontology.ontology_id} ontology...")

            if not Path(ontology_path).exists():
                raise ValueError(f"Ontology file not found: {ontology_path}")

            ontology.load(str(ontology_path))
            # ontology.load()

            data: OntologyData = ontology.extract()
            ontology.build_graph()
            analyzer = self.analyzer_class()
            metrics: OntologyMetrics = analyzer(ontology)

            end_time = time.time()
            processing_time = end_time - start_time

            self.all_metrics[ontology.ontology_id] = {
                "metrics": metrics,
                "ontology_id": ontology.ontology_id,
                "ontology_full_name": ontology.ontology_full_name,
                "domain": ontology.domain,
                "processing_time": processing_time
            }

            self._save_datasets(data, ontology)

            self._generate_documentation(ontology, metrics)

            logger.info(f"Successfully processed {ontology.ontology_id} ontology")
            return metrics
        except Exception as e:
            logger.error(f"Error processing {ontology.ontology_id} ontology: {e}")
            raise

    def _save_datasets(self, data: OntologyData, ontology: BaseOntology) -> None:
        """Save extracted datasets to files"""
        for dataset_type in ['term_typings', 'type_taxonomies', 'type_non_taxonomic_relations']:
            domain_dir = self.datasets_dir / f"{ontology.domain.lower().replace(' ', '_')}"
            domain_dir.mkdir(parents=True, exist_ok=True)

            save_path = domain_dir / f"{ontology.ontology_id.lower().replace(' ', '_')}/{dataset_type}.json"
            io.save_json(data.model_dump()[dataset_type], save_path)

    def _generate_documentation(self, ontology: BaseOntology, metrics: OntologyMetrics):
        """Generate RST documentation from template"""
        context = {
            # Class metadata
            'ontology_name': ontology.ontology_full_name,
            'description': ontology.__doc__.strip() if ontology.__doc__ else "No description available",
            'domain': ontology.domain,
            'category': ontology.category,
            'version': ontology.version,
            'last_updated': ontology.last_updated,
            'creator': ontology.creator,
            'license': ontology.license,
            'format': ontology.format,
            'download_url': ontology.download_url,
            'class_name': ontology.__class__.__name__,

            # Graph metrics
            'total_nodes': metrics.topology.total_nodes,
            'total_edges': metrics.topology.total_edges,
            'root_nodes': metrics.topology.num_root_nodes,
            'leaf_nodes': metrics.topology.num_leaf_nodes,

            # Knowledge coverage
            'num_classes': metrics.topology.num_classes,
            'num_individuals': metrics.topology.num_individuals,
            'num_properties': metrics.topology.num_properties,

            # Hierarchical metrics
            'max_depth': metrics.topology.max_depth,
            'min_depth': metrics.topology.min_depth,
            'avg_depth': f"{metrics.topology.avg_depth:.2f}",
            'depth_variance': f"{metrics.topology.depth_variance:.2f}",

            # Breadth metrics
            'max_breadth': metrics.topology.max_breadth,
            'min_breadth': metrics.topology.min_breadth,
            'avg_breadth': f"{metrics.topology.avg_breadth:.2f}",
            'breadth_variance': f"{metrics.topology.breadth_variance:.2f}",

            # Dataset statistics
            'num_term_types': metrics.dataset.num_term_types,
            'num_taxonomic_relations': metrics.dataset.num_taxonomic_relations,
            'num_non_taxonomic_relations': metrics.dataset.num_non_taxonomic_relations,
            'avg_terms_per_type': f"{metrics.dataset.avg_terms:.2f}"
        }

        content = self.doc_template.render(context)
        domain_dir = self.benchmark_dir / f"{ontology.domain.lower().replace(' ', '_')}"
        domain_dir.mkdir(parents=True, exist_ok=True)
        doc_path = domain_dir / f"{ontology.ontology_id.lower()}.rst"
        with open(doc_path, "w", encoding="utf-8") as f:
            f.write(content)
        logger.info(f"Generated documentation at {doc_path}")

    def export_metrics_to_excel(self):
        """
        Export all collected metrics to an Excel file.
        If the Excel file already exists, read it, update with new metrics, and write it back.
        This preserves metrics for ontologies not processed in the current run.
        """
        if not self.all_metrics:
            logger.warning("No metrics to export. Process at least one ontology first.")
            return

        excel_path = self.metrics_dir / "metrics.xlsx"

        # Try to read existing Excel file if it exists
        existing_df = None
        if excel_path.exists():
            try:
                existing_df = pd.read_excel(excel_path)
                logger.info(f"Read existing metrics from {excel_path}")
            except Exception as e:
                logger.warning(f"Could not read existing metrics file: {e}. Creating a new file.")

        # Create DataFrame from current metrics
        current_rows = []
        for ontology_id, data in self.all_metrics.items():
            metrics = data["metrics"]
            row = {
                "Ontology ID": data["ontology_id"],
                "Ontology Full Name": data["ontology_full_name"],
                "Domain": data["domain"],
                "Ontology Name": metrics.name,
                "Processing Time (s)": data["processing_time"],
                **metrics.topology.dict(),
                **metrics.dataset.dict()
            }
            current_rows.append(row)

        current_df = pd.DataFrame(current_rows)

        # Merge existing and current metrics
        if existing_df is not None and not existing_df.empty:
            # Remove rows for ontologies that are in the current metrics (to be updated)
            current_ontology_ids = current_df["Ontology ID"].tolist()
            existing_df = existing_df[~existing_df["Ontology ID"].isin(current_ontology_ids)]

            # Concatenate existing (minus updated ones) with current
            final_df = pd.concat([existing_df, current_df], ignore_index=True)

            # Sort by Ontology ID for consistency
            final_df = final_df.sort_values("Ontology ID").reset_index(drop=True)

            logger.info(f"Updated metrics for {len(current_ontology_ids)} ontologies, preserved metrics for {len(existing_df)} ontologies")
        else:
            final_df = current_df
            logger.info(f"Created new metrics file with {len(current_rows)} ontologies")

        # Calculate total processing time
        total_processing_time = sum(data.get("processing_time", 0) for data in self.all_metrics.values())
        logger.info(f"Total processing time for all ontologies: {total_processing_time:.2f} seconds")

        if not final_df.empty:
            # Sort the DataFrame by processing time (descending) to see which ontologies took longest
            final_df = final_df.sort_values("Processing Time (s)", ascending=False).reset_index(drop=True)

        # Write to Excel
        final_df.to_excel(excel_path, index=False)
        logger.info(f"Exported metrics to Excel file: {excel_path}")



from ..base import AutoPrompt


class StandardizedPrompting(AutoPrompt):
    def __init__(self, prompt_template: str = None, task: str = None):
        if prompt_template is None:
            if task == "term-typing":
                prompt_template = """
Given a list of types as candidates to be assigned to the term, identify the most probable types.
Return the types only in the form of a list. Do not provide any explanation outside the list.

Term:
{term}

Candidates Types:
{context}

Response:
"""
            elif task == "taxonomy-discovery":
                prompt_template = """
Is {parent} a parent of {child}?
Answer yes/no. Do not explain.
"""
            elif task == "task-non-taxonomic-relations":
                prompt_template = """
What is the relation between {head} and {tail}? Return only the relation type.
"""

        super().__init__(prompt_template)

    def format(self, **kwargs):
        return self.prompt_template.format(**kwargs)



import time
from abc import ABC
from rdflib import RDF, RDFS, OWL
from collections import defaultdict

from .. import logger
from ..base import BaseOntology
from ..data_structure import OntologyData, TopologyMetrics, DatasetMetrics, OntologyMetrics


class Analyzer(ABC):
    """Base class for ontology analysis"""
    def __init__(self):
        pass

    def __call__(self, ontology: BaseOntology) -> OntologyMetrics:
        """Perform complete analysis of ontology"""
        return OntologyMetrics(
            name=ontology.__class__.__name__,
            topology=self.compute_topology_metrics(ontology),
            dataset=self.compute_dataset_metrics(ontology)
        )

    @staticmethod
    def compute_topology_metrics(ontology: BaseOntology) -> TopologyMetrics:
        """Compute comprehensive topology metrics for the ontology."""
        logger.info("Starting topology metrics computation")
        start_time = time.time()
        graph = ontology.nx_graph

        if graph.number_of_nodes() == 0:
            return TopologyMetrics(
                total_nodes=0, total_edges=0,
                num_classes=0, num_properties=0, num_individuals=0,
                max_depth=0, min_depth=0, avg_depth=0, depth_variance=0,
                max_breadth=0, min_breadth=0, avg_breadth=0, breadth_variance=0,
                num_root_nodes=0, num_leaf_nodes=0
            )

        # Depth calculation with multiple root support
        root_nodes = [node for node in graph.nodes() if graph.in_degree(node) == 0]
        leaf_nodes = [node for node in graph.nodes() if graph.out_degree(node) == 0]
        depths = {}
        for root in root_nodes:
            current_visited = {root: 0}
            queue = [root]
            while queue:
                node = queue.pop(0)
                current_depth = current_visited[node]
                if node not in depths or current_depth < depths[node]:
                    depths[node] = current_depth
                for child in graph.successors(node):
                    child_depth = current_depth + 1
                    if child not in current_visited or child_depth < current_visited.get(child, float('inf')):
                        current_visited[child] = child_depth
                        queue.append(child)
            for node, depth in current_visited.items():
                if node not in depths or depth < depths[node]:
                    depths[node] = depth

        # Depth metrics
        depth_values = list(depths.values()) if depths else []
        max_depth = max(depth_values) if depth_values else 0
        min_depth = min(depth_values) if depth_values else 0
        avg_depth = sum(depth_values)/len(depth_values) if depth_values else 0.0
        depth_variance = sum((x - avg_depth)**2 for x in depth_values)/len(depth_values) if depth_values else 0.0

        # Breadth metrics
        depth_groups = defaultdict(list)
        for node, depth in depths.items():
            depth_groups[depth].append(node)
        breadth_values = [len(nodes) for nodes in depth_groups.values()]
        max_breadth = max(breadth_values) if breadth_values else 0
        min_breadth = min(breadth_values) if breadth_values else 0
        avg_breadth = sum(breadth_values)/len(breadth_values) if breadth_values else 0.0
        breadth_variance = sum((x - avg_breadth)**2 for x in breadth_values)/len(breadth_values) if breadth_values else 0.0

        # Knowledge coverage metrics
        classes = set(ontology.rdf_graph.subjects(RDF.type, RDFS.Class)) | \
                  set(ontology.rdf_graph.subjects(RDF.type, OWL.Class))
        num_classes = len(classes)

        properties = set(ontology.rdf_graph.subjects(RDF.type, OWL.ObjectProperty)) | \
                     set(ontology.rdf_graph.subjects(RDF.type, OWL.DatatypeProperty))
        num_properties = len(properties)

        individuals = set()
        for s in ontology.rdf_graph.subjects(RDF.type, None):
            for o in ontology.rdf_graph.objects(s, RDF.type):
                if o in classes:
                    individuals.add(s)
                    break
        num_individuals = len(individuals)

        metrics = TopologyMetrics(
            total_nodes=graph.number_of_nodes(),
            total_edges=graph.number_of_edges(),
            num_classes=num_classes,
            num_properties=num_properties,
            num_individuals=num_individuals,
            max_depth=max_depth,
            min_depth=min_depth,
            avg_depth=avg_depth,
            depth_variance=depth_variance,
            max_breadth=max_breadth,
            min_breadth=min_breadth,
            avg_breadth=avg_breadth,
            breadth_variance=breadth_variance,
            num_root_nodes=len(root_nodes),
            num_leaf_nodes=len(leaf_nodes)
        )

        logger.info(f"Completed topology metrics computation in {time.time() - start_time:.2f} seconds")

        return metrics

    @staticmethod
    def compute_dataset_metrics(ontology: BaseOntology) -> DatasetMetrics:
        """Compute metrics for generated datasets"""
        data: OntologyData = ontology.extract()

        term_typings = data.term_typings
        taxonomies = data.type_taxonomies.taxonomies
        non_taxonomic = data.type_non_taxonomic_relations.non_taxonomies

        class_counts = {}
        for term_type in term_typings:
            for type_name in term_type.types:
                class_counts[type_name] = class_counts.get(type_name, 0) + 1

        avg_terms_per_type = len(term_typings) / len(class_counts) if class_counts else 0

        dataset_metrics = DatasetMetrics(
            num_term_types=len(term_typings),
            num_taxonomic_relations=len(taxonomies),
            num_non_taxonomic_relations=len(non_taxonomic),
            avg_terms=avg_terms_per_type
        )
        return dataset_metrics



import numpy as np
from typing import List, Dict, Any


def aggregate_metrics(results: List[Dict[str, Any]], task: str) -> Dict[str, float]:
    """
    Aggregate metrics across multiple examples
    """
    metrics = {}

    if not results:
        return metrics

    if task == "term-typing":
        metrics['avg_precision_score'] = np.mean([r['precision_score'] for r in results])
        metrics['avg_recall_score'] = np.mean([r['recall_score'] for r in results])
        metrics['avg_f1_score'] = np.mean([r['f1_score'] for r in results])
        metrics['avg_exact_match'] = np.mean([r['exact_match'] for r in results])

    elif task == "taxonomy-discovery":
        metrics['avg_accuracy'] = np.mean([r['accuracy'] for r in results])
        metrics['avg_f1_score'] = np.mean([r['f1_score'] for r in results])

    elif task == "non-taxonomy-discovery":
        metrics['avg_exact_match'] = np.mean([r['exact_match'] for r in results])
        metrics['avg_similarity'] = np.mean([r['similarity_score'] for r in results])

    metrics['num_samples'] = len(results)

    return metrics



from ..base import BaseOntology


class GoodRelations(BaseOntology):
    """
    GoodRelations is a standardized vocabulary (also known as "schema", "data dictionary",
    or "ontology") for product, price, store, and company data that can (1) be embedded
    into existing static and dynamic Web pages and that (2) can be processed by other computers.
    This increases the visibility of your products and services in the latest generation
    of search engines, recommender systems, and other novel applications.
    """
    ontology_id = "GoodRelations"
    ontology_full_name = "Good Relations Language Reference (GoodRelations)"
    domain = "Finance"
    category = "E-commerce"
    version = "1.0"
    last_updated = "2011-10-01"
    creator = "Martin Hepp"
    license = "Creative Commons 3.0"
    format = "OWL"
    download_url = "https://www.heppnetz.de/ontologies/goodrelations/v1"



from pathlib import Path
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd


def plot_model_comparison(summary_df, path: Path):
    """Plot comparison of model performance"""
    plt.figure(figsize=(10, max(6, len(summary_df) * 0.8)))

    # Create a more readable format for model names
    summary_df['Model'] = summary_df.apply(
        lambda x: f"{x['LLM Model'].split('/')[-1]} + {x['Retriever Model'].split('/')[-1]}",
        axis=1
    )

    plot_df = summary_df.sort_values('F1 Score')

    sns.barplot(x='F1 Score', y='Model', data=plot_df, color='blue', label='F1 Score')
    plt.title('Model Performance (F1 Score)')
    plt.xlabel('Score')
    plt.tight_layout()
    plt.savefig(path / "model_f1_comparison.png")

    # Create separate plots for precision and recall
    plt.figure(figsize=(10, max(6, len(summary_df) * 0.8)))
    plot_df_melted = pd.melt(
        plot_df,
        id_vars=['Model'],
        value_vars=['Precision Score', 'Recall Score', 'F1 Score'],
        var_name='Metric', value_name='Score'
    )
    sns.barplot(x='Score', y='Model', hue='Metric', data=plot_df_melted)
    plt.title('Model Performance Metrics')
    plt.tight_layout()
    plt.savefig(path / "model_metrics_comparison.png")


def plot_type_analysis(results, path):
    """Plot analysis of performance by type"""
    path.mkdir(exist_ok=True, parents=True)

    type_metrics = {}

    # Collect performance metrics per type
    for result in results:
        for entry in result['detailed_results']:
            for type_name in entry['ground_truth']:
                if type_name not in type_metrics:
                    type_metrics[type_name] = {'correct': 0, 'total': 0}

                type_metrics[type_name]['total'] += 1
                if type_name in entry['predicted']:
                    type_metrics[type_name]['correct'] += 1

    # Calculate accuracy per type
    type_accuracy = {
        t: metrics['correct'] / metrics['total']
        for t, metrics in type_metrics.items()
        if metrics['total'] > 0
    }

    # Plot type accuracy
    plt.figure(figsize=(12, max(6, len(type_accuracy) * 0.3)))
    types_df = pd.DataFrame({
        'Type': list(type_accuracy.keys()),
        'Accuracy': list(type_accuracy.values()),
        'Count': [type_metrics[t]['total'] for t in type_accuracy.keys()]
    }).sort_values('Accuracy')

    # Color bars by frequency
    colors = plt.cm.viridis(types_df['Count'] / types_df['Count'].max())

    plt.barh(types_df['Type'], types_df['Accuracy'], color=colors)
    plt.xlabel('Accuracy')
    plt.ylabel('Type')
    plt.title('Prediction Accuracy by Type')
    plt.xlim(0, 1)
    plt.tight_layout()
    plt.savefig(path / "type_accuracy.png")

    # Add a color bar to show frequency
    plt.figure(figsize=(12, max(6, len(type_accuracy) * 0.3)))
    scatter = plt.scatter(
        types_df['Accuracy'],
        range(len(types_df)),
        c=types_df['Count'],
        cmap='viridis',
        s=100
    )
    plt.colorbar(scatter, label='# Occurrences')
    plt.yticks(range(len(types_df)), types_df['Type'])
    plt.xlabel('Accuracy')
    plt.title('Type Accuracy with Frequency')
    plt.xlim(0, 1)
    plt.grid(axis='x', linestyle='--', alpha=0.7)
    plt.tight_layout()
    plt.savefig(path / "type_accuracy_with_freq.png")


def plot_precision_recall_distribution(results, path):
    """Plot distribution of precision and recall scores"""
    path.mkdir(exist_ok=True, parents=True)

    all_entries = []
    for result in results:
        model_name = f"{result['llm'].split('/')[-1]} + {result['retriever'].split('/')[-1]}"
        for entry in result['detailed_results']:
            entry['model'] = model_name
            all_entries.append(entry)

    entries_df = pd.DataFrame(all_entries)

    # Precision-Recall scatter plot
    plt.figure(figsize=(10, 8))
    sns.scatterplot(
        data=entries_df,
        x='precision_score',
        y='recall_score',
        hue='model',
        size='f1_score',
        sizes=(20, 200),
        alpha=0.7
    )
    plt.title('Precision Score vs Recall Score for Each Example')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.xlabel('Precision Score')
    plt.ylabel('Recall Score')
    plt.tight_layout()
    plt.savefig(path / "precision_recall_scatter.png")

    # F1 score distribution
    plt.figure(figsize=(10, 6))
    sns.histplot(entries_df, x='f1_score', hue='model', multiple='dodge', bins=10)
    plt.title('Distribution of F1 Scores')
    plt.xlabel('F1 Score')
    plt.ylabel('Count')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.tight_layout()
    plt.savefig(path / "f1_distribution.png")



from abc import ABC
from typing import Any, List


class AutoLearner(ABC):

    def __init__(self, **kwargs):
        pass

    def load(self, **kwargs):
        pass

    def fit(self, train_data: Any, task: str):
        pass

    def predict(self, eval_data: Any, task: str) -> Any:
        pass

    def fit_predict(self, train_data: Any, eval_data: Any, task: str) -> Any:
        self.fit(train_data, task)
        predicts = self.predict(eval_data, task)
        return predicts


class AutoLLM(ABC):
    def __init__(self):
        self.model = None
        self.tokenizer = None

    def load(self, model_id: str):
        pass

    def generate(self, inputs: List[Any]) -> List[Any]:
        pass


class AutoRetriever(ABC):
    def __init__(self):
        self.model = None

    def load(self, model_id: str):
        pass

    def index(self, inputs: List[Any]):
        pass

    def retrieve(self, inputs: List[Any], top_k: int) -> List[Any]:
        pass


class AutoPrompt(ABC):
    def __init__(self, prompt_template: str):
        self.prompt_template = prompt_template

    def format(self, **kwargs):
        pass



import logging
from pathlib import Path
from typing import Dict, List, Optional, Union, Tuple, Any

from .base import AutoLearner, AutoPrompt
from .evaluation import calculate_term_typing_metrics, aggregate_metrics, calculate_taxonomy_metrics, \
    calculate_non_taxonomy_metrics
from .evaluation.visualisations import plot_precision_recall_distribution
from .learner import BERTRetrieverLearner, AutoLearnerLLM, AutoRAGLearner, StandardizedPrompting

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LearnerPipeline:
    """
    Orchestrates the ontology learning pipeline with configuration for
    retriever-only, LLM-only, or combined learning approaches.
    """
    def __init__(self,
                 task: str,
                 learner: Optional[AutoLearner] = None,
                 prompting: Optional[AutoPrompt] = None,
                 retriever: Optional[Any] = None,
                 llm: Optional[Any] = None,
                 retriever_id: Optional[str] = None,
                 llm_id: Optional[str] = None,
                 hf_token: Optional[str] = None):
        """
        Initialize the learning pipeline with flexible configuration.
        """
        self.task = task
        self.results = []
        self.metrics = {}
        self.model_info = {"task": task}

        if retriever_id:
            self.model_info["retriever"] = retriever_id
        if llm_id:
            self.model_info["llm"] = llm_id

        # Case 1: Use pre-configured learner if provided
        if learner is not None:
            self.learner = learner
            if prompting is not None:
                self.prompting = prompting
            return

        # Case 2: Configure components based on what's provided
        if prompting is None:
            self.prompting = StandardizedPrompting(task=task)
        else:
            self.prompting = prompting

        # Case 3: Create retriever if needed
        if retriever is None and retriever_id is not None:
            retriever = BERTRetrieverLearner()
            if retriever_id:
                retriever.load(retriever_id)

        # Case 4: Create LLM if needed
        if llm is None and llm_id is not None:
            llm = AutoLearnerLLM(token=hf_token)
            if llm_id:
                llm.load(llm_id)

        if retriever is not None and llm is not None:
            # RAG-based learner (retriever + LLM)
            self.learner = AutoRAGLearner(retriever, llm, self.prompting)
        elif retriever is not None:
            # Retriever-only learner
            self.learner = retriever
        elif llm is not None:
            # LLM-only learner
            self.learner = llm
        else:
            raise ValueError("At least one of learner, retriever, or llm must be provided")

    def fit(self, train_data, top_k: int = 5) -> 'LearnerPipeline':
        """Train the learner on the given data"""
        self.learner.fit(train_data=train_data, task=self.task)
        return self

    def predict(self, test_data, limit: int = 10) -> List[Dict]:
        """Predict on test data and return results with metrics"""
        results = []
        if self.task == "term-typing":
            test_subset = test_data.term_typings[:limit]
            for typing in test_subset:
                term = typing.term
                ground_truth = typing.types
                predicted = self.learner.predict(term, task=self.task)
                metrics = calculate_term_typing_metrics(predicted, ground_truth)
                results.append({
                    'term': term,
                    'ground_truth': ground_truth,
                    'predicted': predicted,
                    **metrics
                })
        elif self.task == "taxonomy-discovery":
            test_subset = test_data.type_taxonomies.taxonomies[:limit]
            for relation in test_subset:
                parent = relation.parent
                child = relation.child
                ground_truth = "is-a"  # Always "is-a" for taxonomic relations
                predicted = self.learner.predict((parent, child), task=self.task)
                metrics = calculate_taxonomy_metrics(predicted, ground_truth)
                results.append({
                    'parent': parent,
                    'child': child,
                    'ground_truth': ground_truth,
                    'predicted': predicted[0],
                    **metrics
                })
        elif self.task == "task-non-taxonomic-relations":
            test_subset = test_data.type_non_taxonomic_relations.non_taxonomies[:limit]
            for relation in test_subset:
                head = relation.head
                tail = relation.tail
                ground_truth = relation.relation
                predicted = self.learner.predict((head, tail), task=self.task)
                metrics = calculate_non_taxonomy_metrics(predicted, ground_truth)
                results.append({
                    'head': head,
                    'tail': tail,
                    'ground_truth': ground_truth,
                    'predicted': predicted[0],
                    **metrics
                })

        self.results = results
        return results

    def evaluate(self, output_dir: Optional[Union[str, Path]] = None) -> Dict:
        """Evaluate results and generate metrics"""
        if not self.results:
            logger.warning("No prediction results available for evaluation")
            return {}

        self.metrics = aggregate_metrics(self.results, self.task)

        if output_dir:
            output_dir = Path(output_dir)
            output_dir.mkdir(exist_ok=True, parents=True)

            all_results = [{
                **self.model_info,
                'metrics': self.metrics,
                'detailed_results': self.results
            }]

            plot_precision_recall_distribution(all_results, output_dir)
            logger.info(f"Results visualization saved to {output_dir}")

        return self.metrics

    def fit_predict(self,
                    train_data,
                    test_data,
                    top_k: int = 5,
                    test_limit: int = 10) -> List[Dict]:
        """Run fit and predict steps (sklearn-style)"""
        self.fit(train_data, top_k)
        return self.predict(test_data, limit=test_limit)

    def fit_predict_evaluate(self,
                             train_data,
                             test_data,
                             top_k: int = 5,
                             test_limit: int = 10,
                             output_dir: Optional[Union[str, Path]] = None) -> Tuple[List[Dict], Dict]:
        """Run fit, predict and evaluate steps in one call"""
        self.fit(train_data, top_k)
        results = self.predict(test_data, limit=test_limit)
        metrics = self.evaluate(output_dir)
        return results, metrics



from typing import List, Dict
from difflib import SequenceMatcher


def calculate_term_typing_metrics(predicted: List[str], ground_truth: List[str]) -> Dict[str, float]:
    """Calculate metrics for term typing task"""
    predicted_set = set(predicted)
    ground_truth_set = set(ground_truth)

    precision_score = len(predicted_set.intersection(ground_truth_set)) / len(predicted_set) if predicted_set else 0
    recall_score = len(predicted_set.intersection(ground_truth_set)) / len(ground_truth_set) if ground_truth_set else 0
    f1_score = 2 * (precision_score * recall_score) / (precision_score + recall_score) if (precision_score + recall_score) > 0 else 0
    exact_match = 1.0 if predicted_set == ground_truth_set else 0.0

    return {
        'precision_score': precision_score,
        'recall_score': recall_score,
        'f1_score': f1_score,
        'exact_match': exact_match
    }


def calculate_taxonomy_metrics(predicted: bool, ground_truth: bool) -> Dict[str, float]:
    """Calculate metrics for taxonomy relation prediction"""
    accuracy = 1.0 if predicted == ground_truth else 0.0

    tp = 1.0 if predicted and ground_truth else 0.0
    fp = 1.0 if predicted and not ground_truth else 0.0
    fn = 1.0 if not predicted and ground_truth else 0.0

    precision_score = tp / (tp + fp) if (tp + fp) > 0 else 0
    recall_score = tp / (tp + fn) if (tp + fn) > 0 else 0
    f1_score = 2 * (precision_score * recall_score) / (precision_score + recall_score) \
        if (precision_score + recall_score) > 0 else 0

    return {
        'accuracy': accuracy,
        'f1_score': f1_score
    }


def calculate_non_taxonomy_metrics(predicted: str, ground_truth: str) -> Dict[str, float]:
    """Calculate metrics for non-taxonomy relation prediction"""
    exact_match = 1.0 if predicted.lower() == ground_truth.lower() else 0.0
    similarity_score = SequenceMatcher(None, predicted.lower(), ground_truth.lower()).ratio()

    return {
        'exact_match': exact_match,
        'similarity_score': similarity_score
    }



from ..base import AutoRetriever
from typing import Any, List
import torch.nn.functional as F
from sentence_transformers import SentenceTransformer


class BERTRetrieverLearner(AutoRetriever):
    def __init__(self):
        super().__init__()
        self.embedding_model = None
        self.documents = []
        self.embeddings = None

    def load(self, model_id: str = "sentence-transformers/all-MiniLM-L6-v2"):
        self.embedding_model = SentenceTransformer(model_id)

    def index(self, inputs: List[Any]):
        self.documents = inputs
        self.embeddings = self.embedding_model.encode(inputs, convert_to_tensor=True)

    def retrieve(self, query: str, top_k: int = 5) -> List[Any]:
        if self.embeddings is None:
            return []

        query_embedding = self.embedding_model.encode(query, convert_to_tensor=True)

        similarities = F.cosine_similarity(
            query_embedding.unsqueeze(0),
            self.embeddings
        )

        top_k = min(top_k, len(self.documents))

        indices = similarities.topk(top_k).indices.tolist()

        return [self.documents[i] for i in indices]


class NGramRetrieverLearner(AutoRetriever):
    def __init__(self):
        super().__init__()
        self.documents = []
        self.terms = []

    def load(self, model_id=None):
        # No model to load for NGrams
        pass

    def index(self, inputs):
        self.documents = inputs
        self.terms = [self._extract_terms(doc) for doc in inputs]

    def retrieve(self, query, top_k=5):
        query_ngrams = self._extract_terms(query)
        scores = []

        for idx, doc_ngrams in enumerate(self.terms):
            intersection = len(set(query_ngrams) & set(doc_ngrams))
            union = len(set(query_ngrams) | set(doc_ngrams))
            similarity = intersection / union if union > 0 else 0
            scores.append((idx, similarity))

        # Sort by similarity
        scores.sort(key=lambda x: x[1], reverse=True)
        top_indices = [idx for idx, _ in scores[:top_k]]

        return [self.documents[idx] for idx in top_indices]

    @staticmethod
    def _extract_terms(text, n=3):
        # Extract character-level n-grams
        return [text[i:i+n] for i in range(len(text)-n+1)]



from typing import Dict, Union
from typing import List
from uuid import uuid4
from pydantic import BaseModel, Field


class Term(BaseModel):
    """Schema for a term extracted from documents"""
    term: str = Field(..., min_length=1)
    extraction_method: str = Field("abstractive", description="extractive/abstractive")


class DocumentReference(BaseModel):
    """Reference to a document with extraction method"""
    doc_id: Union[int, str]
    extraction_method: str


class Document(BaseModel):
    """Schema for a document"""
    id: Union[int, str]
    title: str
    text: str


class TextualData(BaseModel):
    """Schema for textual data from a single split"""
    terms: List[Term] = []
    documents: List[Document] = []
    term2documents: Dict[str, List[DocumentReference]] = {}


class TermTyping(BaseModel):
    """
    Schema for term typing (Task A)
    """
    ID: str = Field(default_factory=lambda: f"TT_{str(uuid4())[:8]}", description="Unique identifier for the term")
    term: str = Field(..., description="The term being typed")
    types: List[str] = Field(..., description="List of types assigned to the term")


class TaxonomicRelation(BaseModel):
    """
    Schema for taxonomy relations (Task B)
    """
    ID: str = Field(default_factory=lambda: f"TR_{str(uuid4())[:8]}", description="Unique identifier for the relation")
    parent: str = Field(..., description="First term in taxonomy relation")
    child: str = Field(..., description="Second term in taxonomy relation")


class NonTaxonomicRelation(BaseModel):
    """
    Schema for non-taxonomic relations (Task C)
    """
    ID: str = Field(default_factory=lambda: f"NR_{str(uuid4())[:8]}", description="Unique identifier for the relation")
    head: str = Field(..., description="Head term in the relation")
    tail: str = Field(..., description="Tail term in the relation")
    relation: str = Field(..., description="Type of relation")


class TypeTaxonomies(BaseModel):
    """
    Schema for taxonomy information
    """
    types: List[str] = Field(..., description="List of types in the taxonomy")
    taxonomies: List[TaxonomicRelation] = Field(..., description="List of taxonomic relations")


class NonTaxonomicRelations(BaseModel):
    """
    Schema for non-taxonomic relation information
    """
    types: List[str] = Field(..., description="List of types involved in relations")
    relations: List[str] = Field(..., description="List of relation types")
    non_taxonomies: List[NonTaxonomicRelation] = Field(..., description="List of non-taxonomic relations")


class OntologyData(BaseModel):
    """
    Schema for complete ontology data
    """
    term_typings: List[TermTyping] = Field(..., description="List of term typing entries")
    type_taxonomies: TypeTaxonomies = Field(..., description="Taxonomy information")
    type_non_taxonomic_relations: NonTaxonomicRelations = Field(..., description="Non-taxonomic relation information")

class PseudoSentence(BaseModel):
    id: str
    pseudo_sentences: List[str]
    terms: List[str]
    types: List[str]


class SyntheticText2OntoData(BaseModel):
    """
    Schema for synthetic text2onto generator
    """
    child_to_parent: Dict[str, List[str]] = Field(..., description="Mapping from child terms to their parent terms")
    pseudo_sentences: List[PseudoSentence] = Field(..., description="List of pseudo sentence batches with metadata")
    generated_docs: List[Document] = Field(..., description="Generated documents from pseudo sentences")



from ..base.ontology import BaseOntology


class AGROVOC(BaseOntology):
    """
    AGROVOC is a relevant Linked Open Data set about agriculture available for public use and facilitates
    access and visibility of data across domains and languages. It offers a structured collection of agricultural concepts,
    terms, definitions and relationships which are used to unambiguously identify resources, allowing standardized
    indexing processes and making searches more efficient.
    """
    ontology_id = "AGROVOC"
    ontology_full_name = "AGROVOC Multilingual Thesaurus (AGROVOC)"
    domain = "Agriculture"
    category = "Agricultural Knowledge"
    version = "2024-04"
    last_updated = "August 12, 2024"
    creator = "Food and Agriculture Organization of the United Nations"
    license = "Creative Commons 4.0"
    format = "RDF"
    download_url = "https://agroportal.lirmm.fr/ontologies/AGROVOC"


class ATOL(BaseOntology):
    """
    ATOL (Animal Trait Ontology for Livestock) is an ontology of characteristics defining phenotypes of livestock
    in their environment (EOL). ATOL aims to:
    - provide a reference ontology of phenotypic traits of farm animals for the international scientific and educational
    - communities, farmers, etc.;
    - deliver this reference ontology in a language which can be used by computers in order to support database management,
    semantic analysis and modeling;
    - represent traits as generic as possible for livestock vertebrates;
    - make the ATOL ontology as operational as possible and closely related to measurement techniques;
    - structure the ontology in relation to animal production.
    """
    ontology_id = "ATOL"
    ontology_full_name = "Animal Trait Ontology for Livestock (ATOL)"
    domain = "Agriculture"
    category = "Animal Science"
    version = "6.0"
    last_updated = "May 11, 2020"
    creator = "INRAE, France"
    license = "Creative Commons 4.0"
    format = "OWL"
    download_url = "https://bioportal.bioontology.org/ontologies/ATOL"


class FoodOn(BaseOntology):
    """
    FoodOn, the food ontology, contains vocabulary for naming food materials and their anatomical and taxonomic origins,
    from raw harvested food to processed food products, for humans and domesticated animals.
    It provides a neutral and ontology-driven standard for government agencies,
    industry, nonprofits and consumers to name and reference food products and their components
    throughout the food supply chain.
    """
    ontology_id = "FoodOn"
    ontology_full_name = "Food Ontology (FoodON)"
    domain = "Agriculture"
    category = "Diet, Metabolomics, and Nutrition"
    version = None
    last_updated = "2025-01-16"
    creator = None
    license = "Creative Commons 4.0"
    format = "OWL"
    download_url = "http://purl.obolibrary.org/obo/foodon.owl"


class PO(BaseOntology):
    """
    The Plant Ontology (PO) is a structured vocabulary and database resource that links plant anatomy,
    morphology and growth and development to plant genomics data.
    """
    ontology_id = "PO"
    ontology_full_name = "Plant Ontology (PO)"
    domain = "Agriculture"
    category = "Plant Anatomy, Morphology, Growth and Development"
    version = None
    last_updated = None
    creator = None
    license = "Creative Commons 4.0"
    format = "OWL"
    download_url = "https://github.com/Planteome/plant-ontology"
